!+
! Module: omlcbuild_mod
!
! Authors:  Tim Harries - MSSL
!          
! Description: Constructs an OM OSW FITS source timeseries from source and background rates files produced by using evselect
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!
! History:
! 
!    15.05.2001 VNY -  timeStart parameter was corrected (TIMEZERO replaced with TSTART)
!    15.05.2001 VNY -  the output table title has been changed from RATES to RATE (necessary for lcplot programme) 
!    04.06.2001 VNY  - wdxSet input parameter is added (it is necessary 
!                      for count rates correction )
!    05.06.2001 VNY - sourceList input parameter is added (for count rates correction)
!    06.06.2001 VNY - makes correction for count losses
!    26.06.2001 VNY - CorrectedCountRate function has been changed
!    06.07.2001 VNY - Corrected the background estimation
!    08.08.2001 VNY - correction of the counts detected within the source radius is made
!                     now taking into account possible lost of photons at the edge of the OSW
!                     and using CAL_omGetPSFmap routine 
!    13.08.2001 VNY - Test with just wider PSF : ((radius-6)**2)/36
!    15.08.2001 VNY - the use of Tracking history file in order to compensate shifts of the source
!                     during the exposure (new parameter is used with the name of the TH-file: thxSet)
!    28.08.2001 VNY - various test with the Tracking History file correcting shifts within one exposure
!    29.09.2001 VNY - Correcting the source count rates when it is near the edge of the field of view
!    30.08.2001 VNY - Average magnitude is inserted to the header of the output data set file istead of
!                     the magnitude column 
!    05.09.2001 VNY - Averaging TH shifts instead of interpolation
!    06.09.2001 VNY - PSF_FRAC parameter has been added to the output table attributes (means the fraction
!                     of the image, which lies within the OSW, just to indicate whether or not the extracted
!                     light curve is affected by the proximity of the source to the edge of the field of view
!    19.09.2001 VNY - checking the background counts for extremely high values
!    20.09.2001 VNY - checkin the source counts for extremely high values
!    21.09.2001 VNY - median filtration of the output time series
!    24.09.2001 VNY - The use of the Tracking History file for the FOV-proximity correction has been removed.
!                     Respectively, the txhset input parameter has been removed as well.
!    18.10.2001 VNY - unused variables removed 
!    25.10.2001 VNY - initialisation of background rate errors array has been introduced (SPR-2611)
!    29.10.2001 VNY - initialisation of countRateErrorMinus and countrateErrorPlus parameters 
!                     has been introduced to the CorrectedCountRate function
!    30.10.2001     - rectified the calculation of rateSourceErrOutArray
!                     (rateBackOutArray was replaced with rateBAckErrOutArray)
!    31.10.2001 VNY - rectified the correction for the PSF fraction outside the OSW
!    01.11.2001 VNY - rectified count rates for the case if the COINCIDENCE table missed in OM_PHOTONAT_XXXX.CCF
!                     (to avoid the Illegal Count Rate messages)
!    13.11.2001 VNY - background is being removed from the source counts before applying the PSF fraction
!                     correction. Then it is restored in order to calculate correctly the 
!                     Coincidence Lost Correction. And after that, the background is 
!                     subtracted again, but at this time from the area defined by CAL_Aperture_Radius.
!                     (SPR-2629) 
!    19.11.2001 VNY - 3 additional keywords have been added to the binary table extension (SRRADIUS, CLRADIUS
!                     and PHRADIUS, the source extraction radius, the coincidence loss correction radius
!                     and the photometry radius). 
!                     Also the parameter VIGNET has been replaced with PSF_FRAC keyword, because vignetting
!                     is defined in OGIP as a geometrical parameter (fraction of the area), but the task
!                     calculates the fraction of the PSF, not of the area. Respectively, PSF_CAPP keyword
!                     has been added, which means the the correction for PSF_FRAC was applied
!   06.12.2001 VNY -  background counts scaling has been corrected in order to get uniform lightcurve
!                     level for different lightcurve extraction areas (fixing SSC-SPR-2668).
!                     New parameter (binBpe) has been added to getFrameTime routine, according to
!                     the changes made by CB in this routine
!   07.12.2001 VNY -  VIGNET parameter is set to 1.0 (no vignetting) in order to permit lcurve to build 
!                     properly the graph
!   10.12.2001 VNY -  improved background counts scaling in order to get uniform lightcurve level
!                     for different lightcurve extraction areas (fixing SSC-SPR-2668).
!                     New parameter (binBpe) has been added to the getFrameTime call,
!                     in accordance with recent changes made by CB in the getFrameTime routine.
!                     VIGNET keyword in the output file header is set to 1.0 (no
!                     vignetting) in order to permit the proper use of this file by LCURVE task.   
!   11.12.2001 VNY -  fine agjustment of the PSF fraction calculation has been made: avoiding circles 
!                     from the neighbour sources are taken into account in accordance with recent changes
!                     in the OMREGION task. 
!   12.12.2001 VNY -  the avoiding circle is variable (because its radius it taken from omdetect). This
!                     could cause a variability of the source area from exposure to exposure. This
!                     could noticably affect the output light curve.
!   17.12.2001 VNY -  TIME column has been added to the output light curve FITS file
!   07.01.2002 SRR -  changed TIMEZERO value to 0.d0 in order to get values of time in the TIME column
!                     referred directly to MJDREF.
!   17.01.2002 VNY -  protected against the absence of data in the background RATE file
!   23.01.2002 VNY -  allocation of timeBack and timeSource arrays has been changed to satisfy dlocate routine:
!                     in dlocate these arrays are supposed to be starting from 1, but a 0 value is 
!                     also read by it, that caused a subscript crash when the array has only one component
!   25.01.2002 VNY -  Scaling background counts to the source area has been improved. The source area 
!                     is not circular for the sources near the OSW edge. Now the real shape and size of this 
!                     area is taken into account before scaling and subtracting background from the 
!                     source counts (SSC-SPR-2715).   
!   05.02.2002 VNY -  Error section in the description document file has been updated (SSC-SPR-2730)
!
!   08.02.2002 VNY - Negative background counts are zeroed. These could appear as the result of 
!                    interpolation between input and output time bins when there are only a few background
!                    counts in the input data (SSC-SPR-2738). In this case the error bars for the zeroed counts 
!                    are deduced from the background median 
!
!  13.02.2002 VNY -  Background calculation has been improved: Fraction of PSF within the background extraction
!                    area has been taken into account. Background counts are corrected for this PSF fraction before
!                    subtracting background from the source counts. This could be essential for small source
!                    extraction radii (SSC-SPR-2745).
!
!  22.03.2002 VNY - Memory allocation/deallocation has been corrected (SPR-2814) 
!
!  26.03.2002 (VNY) DISTRIBUTION has been changed from ssc to world
!  27.03.2002 (VNY) Protected against 0-size of the input source rate table
!  28.03.2002 (VNY) Additional protection against 0-size of the input source rate table 
!  28.03.2002 (VNY) Corrected a bug in memory deallocation introduced on 22.03.2002
!  29.03.2002 (VNY) Default values of integers are forced to be int32-kind
!  30.03.2002 (VNY) Check of the input source rates table is introduced
!
!  18.04.2002 (VNY) Correction of scaling factors, which makes the task more robust in calculation
!                   of count rates for faint objects (SSC-SPR-2826). 
!                   
!

module omlcbuild_mod

  use dal
  use cal
  use param  
  use errorHandling
  use frameTime_mod
  use datafilter
  use sasdal

  implicit none 
  
  private 

  public omlcbuild

contains 
  
  !---------------------------------------------------------------------
  subroutine omlcbuild
  !---------------------------------------------------------------------

   ! variables
   ! a temp DAL dataset to find area of regions

   type(DataSetT) :: tempSet
   type(TableT) :: tempTable
   type(ColumnT) :: tempColX, tempColY


   integer(Kind=int32), pointer :: xArray(:), yArray(:)
   integer(kind=int32) :: npts, nPix
   integer(kind=int32) :: nx, ny
   integer(kind=int32) :: i,j,k

   real(kind=double) :: areaIn, areaOut, areaIn2, areaOut2, areaIn3, areaOut3
   real(kind=double) :: areaBkg

   real(kind=single) :: bkgExtractionArea
   logical :: outOfAvoid
   real(kind=single) :: dist
   integer(kind=int32) :: ia
   integer(kind=int32) :: nBackAvoid
   real (kind=single), dimension(10) :: xBackAvoid
   real (kind=single), dimension(10) :: yBackAvoid
   real (kind=single), dimension(10) :: rBackAvoid

   ! region and lc file filenames

   character(len=1024) :: sourceReg, backReg, expression
   character(len=1024) :: backRates, sourceRates, sourceList, wdxSet

   ! region file variables

   real(kind=single) :: sourceArea, backArea, detectedArea
   type(DataSetT) :: sourceSet, backSet, sourceRegSet, backRegSet
   type(TableT) :: sourceTable, backTable

!---VNY 15.06.2001
   type(TableT) :: regionTable
   type(ColumnT) :: sourceRadiusCol, sourceXPosCol, sourceYPosCol, sourceShapeCol

   type(TableT) :: backRegionTable
   type(ColumnT) :: backRadiusCol, backXPosCol, backYPosCol, backShapeCol

!---VNY 04.06.2001
   ! Priority Window Data dataset handle
   type(DataSetT) :: wdxDataSetH
   ! Input table handle
   type(TableT) :: inTableH
   integer, parameter :: noColumns = 15
   type(ColumnT), dimension(noColumns) :: inColumnH
   ! Input dataset handle
   type(DataSetT) :: inDataSetH
   real(kind=double)              :: detectionLimitCountRate         ! The detection limit count rate (counts/second)
   real(kind=double)              :: detectionLimitMagnitude         ! The detection limit magnitude
   real(kind=double) :: error = 0.0 
   real(kind=single) :: calApertureRadius = 12.0          ! Radius used for coincidence losses   
   real(kind=single) :: photometryApertureRadius          ! Radius used for converting counts to magnitudes
 
   character(len=8) :: dataMode                                      ! Input file data mode.
   character(len=9) :: filter                                        ! Keywords.
   real(kind=single) :: exposure
   real(kind=double) :: frameTime                                    ! OM CCD frame time.
   real(kind=double) :: deadFrac                                     ! Dead time in units of CCD frame time.
   integer(kind=int32) ccdMode                                       ! OM CCD data mode.
   integer(kind=int32) :: filterIdInt

   logical :: scaleCounts = .false.
   logical :: vignApp
   logical :: backApp
   logical :: deadApp
   logical :: psfApp
!---VNY 17.01.2001
   logical :: bkgRateExists   ! Indicates that background rate table exists in the file
   logical :: bkgRateId      ! Indicates that background and source rates are of the same EXP_ID

!---VNY 27.03.2002
   logical :: srcRateExists ! Indicate that source rate table does exist and is not of zero size

   real(kind=single) :: countsRatio = 1.0
   
   real(kind=single) :: vignet =1.0 ! Vignetting parameter
   real(kind=single) :: psfFrac =0.0
   real(kind=double) :: corrRate                                     ! Corrected count rate.
   real(kind=double) :: corrRateErrPlus                              ! Corrected rate error + (asymmetric).
   real(kind=double) :: corrRateErrMinus                             ! Corrected rate error - (asymmetric).
   real(kind=double) :: corrRateErrPlus1                              ! Corrected rate error + (asymmetric).
   real(kind=double) :: corrRateErrMinus1                             ! Corrected rate error - (asymmetric).
   real(kind=double) :: photcal                          ! ratio between the PSF encircled by the photometry and cal radii
   real(kind=double) :: caldet                           ! ratio between the PSF encircled by the cal and detected radii
   real(kind=double) :: sumIn2, sumIn3, sumOut2, sumOut3
   real(kind=double) :: sumBkg, testSumBkg    ! PSF-fraction within background extraction area

   real(kind=double) :: corrMag
         ! Corrected magnitude
   real(kind=double) :: corrMagErrPlus                               ! Corrected magnitude error + (asymmetric).
   real(kind=double) :: corrMagErrMinus                              ! Corrected magnitude error - (asymmetric).
   logical :: extend
   real(kind=single) :: countFrameRatio = 0.5
   real(kind=single), dimension(:), pointer :: rate                  ! Rates pointer.
   real(kind=single), dimension(:), pointer :: rateErr               ! Rates error pointer.

!--- 14.11.2001
   real(kind=double), dimension(:), pointer :: sourceRA              ! Source Right Ascension
   real(kind=double), dimension(:), pointer :: sourceDEC              ! Source Declination
   real(kind=double), dimension(:), pointer :: sourceLII              ! Source Galactic Longitude
   real(kind=double), dimension(:), pointer :: sourceBII              ! Source Galactic Lattitude
   integer(kind=int8), dimension(:), pointer :: sourceQF              ! Source quality flag
   integer(kind=int8), dimension(:), pointer :: sourceCF              ! Source confusion flag

   real(kind = double), parameter :: PI = 3.141592653589793_double  
!****************************************************
! 07-06-2001 CB Need to read in the source coordinates 
! and aperture radius columns
!****************************************************
   real(kind=single), dimension(:), pointer :: sourceXCoord          ! Source x-coordinate.
   real(kind=single), dimension(:), pointer :: sourceYCoord          ! Source y-coordinate. 
   real(kind=single), dimension(:), pointer :: sourceApertureRadius  ! Source aperture radius
   real(kind=single), dimension(:), pointer :: backGroundRate        ! Source background rate pointer.
   real(kind=single), dimension(:), pointer :: mag                   ! Magnitude pointer.
   real(kind=single), dimension(:), pointer :: magErr                ! Magnitude error pointer.
   integer(kind=int8), dimension(:), pointer :: extended
   integer :: iLoop

   integer(kind=int32) :: nSources                                   ! Number of sources.
!---VNY 29.03.2002
   integer(kind=int32) :: nSource, nSourceRateRows                   ! Number of counts in the RATE table 
!---
   integer(kind=int32) :: nNegative                                  ! Number of -ve count rates.

   ! output variables

   type(DataSetT) :: outSet
   type(TableT) :: outTable
   character(len=1024) :: outFile
 
   ! timeline variables

   real(kind=double), allocatable :: timeSource(:)
   real(kind=double), allocatable :: timeBack(:)
   real(kind=double), allocatable :: timeOut(:)

   ! rates file columns

   type(ColumnT) :: rateSourceCol, rateBackCol
   type(ColumnT) :: rateSourceOutCol
   type(ColumnT) :: rateSourceErrOutCol, rateBackOutCol
   type(ColumnT) :: rateBackErrOutCol
   type(ColumnT) :: timeSourceOutCol

   ! rates file pointers

   real(kind=Single), allocatable :: rateSourceArray(:), rateBackArray(:)
   real(kind=Single), allocatable :: SourceArrayError(:), BackArrayError(:)
   integer(kind=int32), pointer :: iSourceArray(:), iBackArray(:)
   integer(kind=int8) :: bgsubtracted

   real(kind=Single), pointer :: rateSourceOutArray(:)
   real(kind=Single), pointer :: rateSourceErrOutArray(:)
   real(kind=Single), pointer :: rateBackOutArray(:)
   real(kind=Single), pointer :: rateBackErrOutArray(:)
   real(kind=Double), pointer :: timeSourceOutArray(:)

   real(kind=double) :: source, background
   real(kind=double) :: timeStart, bkgTimeStart
   real(kind=double) :: timeEnd
   real(kind=double) :: timeDel, bkgTimeDel
   real(kind=double) :: timeExposure

!---VNY 30.08.2001
   real(kind=double) :: mgAver
   real(kind=double) :: nmgAver
   real(kind=double) :: mgMin
   real(kind=double) :: mgMax
   real(kind=double) :: rateAver

!---VNY 21.09.2001 array for the median filtration
   real(kind=single), dimension(9) :: rateAuxArray 
   real(kind=single) :: rateAux

   integer(kind=int32) :: nBack, nRegion, nBackRegion
   integer(kind=int32) :: srcNo
   real(kind=double) :: sourceErr
   real(kind=double) :: backErr

!---VNY 11.12.2001
   integer :: nAvoidingSources
   real(kind=Single) :: avoidX(1:10)
   real(kind=Single) :: avoidY(1:10)
   real(kind=Single) :: avoidR(1:10)
   logical :: outsideOfAvoidingCircle


!--- VNY 15.06.2001
   real(kind=Single), pointer :: radiusDetectedArray(:)
   real(kind=Single), pointer :: sourceXPosArray(:)
   real(kind=Single), pointer :: sourceYPosArray(:)

   real(kind=Single), pointer :: backRadiusArray(:)
   real(kind=Single), pointer :: backXPosArray(:)
   real(kind=Single), pointer :: backYPosArray(:)

   character(len=20) :: sourceShape
   character(len=20) :: backShape
   character(len=80) :: sourceShapeComment
   character(len=80) :: backShapeComment

!--- V.Ye. 14.05.2001 - origin of the OSW
    character(len=255) :: window0                                   ! Origin of the OSW
    integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1   ! Coordinates of the OSW origin
    character(len=1) :: ch1, ch2, ch3, ch4, ch5

!---VNY 08.08.2001 **************************
!   Variables for checking PSF-mapping routine
    real(kind=single), dimension(:, :), pointer :: psf ! To hold the point-spread function

    integer(kind=int32) :: xRes, yRes

    real(kind=double) :: xPos0, yPos0
    integer(kind=int32) :: ix, iy
    real(kind=single) :: dxLeft, dxRight, dyDown, dyUp
    real(kind=single) psfRadius, xyDistance, xc, yc, sumIn, sumOut, radiusDetected
    real(kind=double) :: PsfFractionOfTotalCounts
    real(kind=single) :: radiusTest

!---VNY 06.12.2001
    logical :: binBpe

!---Test VNY 07.02.2002
    integer(kind=int32) :: backMax
    integer(kind=int32), allocatable :: backDist(:)
    real(kind=double) :: backSum, backSumFound, medianBkg, medianBkgError
    logical :: medianFound
    real(kind=double) :: prop
    integer(kind=int32) :: ja
    real(kind=double) :: ratioBS
! VNY 10.04.2002
    real(kind=double) :: testBkg, testInt, testSrc, testZnam

   ! get parameters

   sourceReg = stringParameter("srcregionset")
   backReg = stringParameter("bkgregionset")
   sourceRates = stringParameter("srcrateset")
   backRates = stringParameter("bkgrateset")
   sourceList = stringParameter("sourcelistset")
   wdxSet = stringParameter("wdxset")
   outFile = stringParameter("outset")

   sourceRegSet = dataSet(TRIM(sourceReg), read)
   backRegSet = dataSet(TRIM(backReg), read)

!---VNY 30.03.2002 Check of the length of the source rate table
   ! get a handle on the source rates file
   sourceSet = dataSet(TRIM(sourceRates), read)   
   srcRateExists=.false.
   nSourceRateRows=0
   if (hasBlock(sourceSet, "RATE")) then
      sourceTable = table(sourceSet,"RATE")
      nSourceRateRows = numberOfRows(sourceTable) ! Number of points in the light curve   
      write(errstr,*) "Length of the Source Rates Table:  ", nSourceRateRows
      call message(AppMsg, 2)
      call release(sourceTable)
   endif
   call release(sourceSet)
!---

!---VNY 05.06.2001 
   ! Read in the OM OSW Source List file
   inDataSetH=dataSet(sourceList, modify)
   !---VNY 17.01.2001 Protection against the absence of the SRCLIST block
   if (hasBlock(inDataSetH, "SRCLIST")) then
      !**************************************************************
      ! Read in the detection limit count rate from the PRIMARY header
      !**************************************************************
      detectionLimitCountRate = real(real32Attribute(inDataSetH, "CRLIMIT"))
           
      ! Get a handle on the source list binary table extension.
      inTableH = table(inDataSetH, "SRCLIST")

      ! Read in the source list table extension keywords.
      dataMode = stringAttribute(inDataSetH, "DATAMODE")
      filter = stringAttribute(inDataSetH, "FILTER")
      exposure = real32Attribute(inDataSetH, "EXPOSURE")

      !---VNY 06.12.2001 
      binBpe= booleanAttribute(inDataSetH, "BINBPE")

      !********************************************************
      ! Read in the OM ODF Priority Window Data Auxiliary file.
      !********************************************************
      wdxDataSetH = dataSet(wdxSet, read)

      !************************************************************************
      ! Get the frame time from the OM ODF Priority Window Data Auxiliary file.
      !************************************************************************

      !---VNY 06.12.2001 New parameter added to getFrameTime: binBpe
      call getFrameTime(wdxDataSetH, frameTime, deadFrac, binBpe)  ! Frame time is in milli-seconds
   
      frameTime = frameTime / 1000.0    ! Convert frame time to seconds

      !********************
      ! Get the ccd mode.
      !********************
      if (trim(dataMode) .eq. "IMAGING") then
         ccdMode = IMAGE
      else
         ccdMode = FAST
      end if

      !**************************************
      ! Get the filter Id integer value
      !**************************************
      filterIdInt = filterIdToInt(filter)

      !*******************
      ! Set the CAL state
      !********************
      call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)

      !*************************************************************
      ! 17th May 2001 - Due to the changes to the photometry in
      ! omdetect, now need to scale the source counts from the 12
      ! pixel aperture radius to that specified by the CAL.
      ! For U, B and V there will be no correction.
      !************************************************************

      photometryApertureRadius = CAL_omGetApertureRadius()

      if (photometryApertureRadius .gt. calApertureRadius) then
         !**********************************************
         ! Compute the correction factor needed to scale  
         ! the source counts from an aperture radius of
         ! 12.0 pixels to the apertureRadius
         !**********************************************
         scaleCounts = .true.
      end if

      !**************************
      ! Release handles on data.
      !**************************
      call release(wdxDataSetH)

      !******************************************************
      ! Calculate the instrumental limiting magnitude from
      ! the detection limit count rate
      !******************************************************
      if (detectionLimitCountRate .gt. 0.0) then      
         corrRate = CorrectedCountRate(detectionLimitCountRate, error, frameTime, &
            deadFrac, .false., corrRateErrMinus, corrRateErrPlus)
         if (corrRate .gt. 0.0) then
            !**********************************************************
            ! May need to scale the counts from 12 to 35 pixels for uv
            !*********************************************************
            if (scaleCounts) then
               corrRate = corrRate * (photometryApertureRadius / calApertureRadius) ** 2
            end if
            call CAL_omPhotoMagnitude(corrRate, detectionLimitMagnitude, corrRateErrMinus, &
                                   corrRateErrMinus, corrMagErrMinus, corrMagErrPlus)
         else
            detectionLimitMagnitude = 0.0d0
         end if
      else
         detectionLimitmagnitude = 0.0
      end if
      call setAttribute(inDataSetH, "MAGLIMIT", real(detectionLimitMagnitude), "", "detection limiting magnitude")

      !******************************************************************
      ! Get the number of rows in the source list binary table extension.
      !******************************************************************
      nSources = numberOfRows(inTableH)

      !---VNY 26.03.2002 Setting a limit to the input rows to 1000 sources
      if (nSources .gt. 0 .and. nSources .lt. 1000) then
         ! Get handles and pointers on the source list binary table extension colums.
         inColumnH(1) = column(inTableH, "RATE", modify)
         rate => real32Data(inColumnH(1))

         inColumnH(2) = column(inTableH, "RATE_ERR", modify)
         rateErr => real32Data(inColumnH(2))

         inColumnH(3) = column(inTableH, "MAG", modify)
         mag => real32Data(inColumnH(3))

         inColumnH(4) = column(inTableH, "MAGERR", modify)
         magErr => real32Data(inColumnH(4))

         inColumnH(5) = column(inTableH, "EFLAG", modify)
         extended => int8Data(inColumnH(5))
     
         inColumnH(6) = column(inTableH, "BACKGROUND_RATE", modify)
         backGroundRate => real32Data(inColumnH(6))

         inColumnH(7) = column(inTableH, "XPOS", read)
         sourceXCoord => real32Data(inColumnH(7))

         inColumnH(8) = column(inTableH, "YPOS", read)
         sourceYCoord => real32Data(inColumnH(8))
   
         inColumnH(9) = column(inTableH, "APERTURE", read)
         sourceApertureRadius => real32Data(inColumnH(9))

         inColumnH(10) = column(inTableH, "RA", read)
         sourceRA => real64Data(inColumnH(10))

         inColumnH(11) = column(inTableH, "DEC", read)
         sourceDEC => real64Data(inColumnH(11))

         inColumnH(12) = column(inTableH, "LII", read)
         sourceLII => real64Data(inColumnH(12))       
       
         inColumnH(13) = column(inTableH, "BII", read)
         sourceBII => real64Data(inColumnH(13))

         inColumnH(14) = column(inTableH, "QFLAG", read)
         sourceQF => int8Data(inColumnH(14))

         inColumnH(15) = column(inTableH, "CFLAG", read)
         sourceCF => int8Data(inColumnH(15))
   
         !******************************************************************
         ! Loop through all the sources  - 
         ! Correct each count-rate for the detector's 
         ! lifetime and ageing, then calculate the magnitude.
         !******************************************************************   
         nNegative = 0   
         iLoop=0
         if (extended(iLoop) .eq. 0.0) then
               extend = .false.
            else
               extend = .true.
            end if

            ! set up a large temporary event-list of 2048x2048 elements
            nx = 2048
            ny = 2048
            tempSet = dataSet("tmp.fts", temp)
            tempTable = addTable(tempSet,"EVENTLIST", nx*ny)
            tempColX = addColumn(tempTable, "X", integer32)
            tempColY = addColumn(tempTable, "Y", integer32)
            xArray => int32data(tempColX)
            yArray => int32data(tempColY)
            npts = 0
            do i = 0, nx-1
               do j = 0, ny-1
                  xArray(npts) = i
                  yArray(npts) = j
                  npts = npts + 1
               enddo
            enddo

            ! filter the temp event-list by the source region, which
            ! gives the number of pixels contained within the source area

            write(expression,'(a,a,a)') "region(",trim(sourceReg),", X,Y)"
            call initFilter(expression)
            nPix = doFilter(tempTable) 
            sourceArea = real(nPix)

            write(errstr,'(a,f6.1)') "Source region area (pixels): ",sourceArea
            call message(AppMsg, 2)

!-- VNY 22.03.2002
            call release(tempColY)
            call release(tempColX)
            call release(tempTable)
!---  
            call release(tempSet)

            ! do the same for the background
            tempSet = dataSet("tmp.fts", temp)
            tempTable = addTable(tempSet,"EVENTLIST", nx*ny)
            tempColX = addColumn(tempTable, "X", integer32)
            tempColY = addColumn(tempTable, "Y", integer32)
            xArray => int32data(tempColX)
            yArray => int32data(tempColY)
            npts = 0
            do i = 0, nx-1
               do j = 0, ny-1
                  xArray(npts) = i
                  yArray(npts) = j
                  npts = npts + 1
               enddo
            enddo
            write(expression,'(a,a,a)') "region(",trim(backReg),", X,Y)"
            call initFilter(expression)
            nPix = doFilter(tempTable)
            backArea = real(nPix)

            write(errstr,'(a,f6.1)') "Background region area (pixels): ",backArea
            call message(AppMsg, 2)
!--VNY 22.03.2002
            call release(tempColY)
            call release(tempColX)
            call release(tempTable)
!---

            call release(tempSet)

            srcNo = int32Attribute(sourceRegSet, "SRCNO")
            timeExposure = real64Attribute(sourceRegSet, "EXPOSURE") 

            ! get a handle on the source rates file
            sourceSet = dataSet(TRIM(sourceRates), read)   

            !--- V.Ye. 14.05.2001 (origin of the OSW)
            window0 = stringAttribute(sourceSet, "DETSEC")

            !--- 15.05.2001 V.Ye.: the origin of the OSW has been added in order to be able to extract time series
            read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                 ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
            
            !---VNY 27.03.2002
            srcRateExists=.false.

            !--- VNY 16.01.2002
            if (hasBlock(sourceSet, "RATE")) then
               sourceTable = table(sourceSet,"RATE")
               nSource = numberOfRows(sourceTable) ! Number of points in the light curve

               !---VNY 26.03.2002 Protection against big values of nSource
               if (nSource .gt. 10000) nSource=9999
               if (nSource .lt. 0) nSource=-nSource

               ! ---VNY 29.03.2002 
               !************************************
               write (errstr, *) "Source RATE table size in the SCRATE file:  ", nSource
               call message(AppMsg, 2)

               !---VNY 27.03.2002 protection against zero size of source rates table
               if (nSource .gt. 2) then 
                  srcRateExists=.true.
               else
                  !---VNY 28.03.2002 additional protection against zero size of source rates table
                  call warning("srcRATE", "Source RATE table is too small in the SCRATE  file")
                  !---VNY 30.03.2002 Stop if nSource is less or equals to zero
                  if (nSourceRateRows .gt. 2) then
                     nSource=nSourceRateRows
                     call warning("corrSrcRate", " Correcting source RATE table length")
                     srcRateExists=.true.
                  else
                     call fatal("dataERR", " Impossible to get correct source rates table length")
                  endif
               endif

               rateSourceCol = column(sourceTable,"COUNTS", read)
               iSourceArray => int32data(rateSourceCol)

               !--- VNY 15.06.2001: reading the source aperture radius detected by OMREGION
               regionTable = table(sourceSet, "REGION")
               sourceShapeCol = column(regionTable, "SHAPE", read)
               sourceRadiusCol = column(regionTable, "R", read)
               sourceXPosCol = column(regionTable, "X", read)
               sourceYPosCol = column(regionTable, "Y", read)
               nRegion = numberOfRows(regionTable)
               radiusDetectedArray => real32data(sourceRadiusCol)
               sourceXPosArray => real32data(sourceXPosCol)
               sourceYPosArray => real32data(sourceYPosCol)

               !---VNY 11.12.2001 fixing number of avoiding sources 
               nAvoidingSources=0
               if (nRegion .gt. 1) then 
                  nAvoidingSources=nRegion-1
               end if

               ! VNY 11.12.2001 Limiting the number of avoiding sources
               if (nAvoidingSources .gt. 10) then 
                  nAvoidingSources=10
               end if

               ! VNY 11.12.2001 Preparing an array of coordinates of avoiding sources relative to
               ! the centre of the target source
               do i=1,nAvoidingSources
                  k=i*50 ! each source is a sub-array of 50 elements
                  avoidX(i)=sourceXPosArray(k)-sourceXPosArray(0)
                  avoidY(i)=sourceYPosArray(k)-sourceYPosArray(0)
                  avoidR(i)=radiusDetectedArray(k)
               end do

               ! Prepare comment on the source region
               sourceShapeComment=" "
               if (nRegion .gt. 0) then
                  sourceShape = stringCell(sourceShapeCol,0)
                  ! Calculate the source shape word length              
                  wordlength: do i=1,20
                     if (ichar(sourceShape(i:i)) .eq. 0) then
                        exit wordlength
                     end if
                  end do wordlength
 
                  write(sourceShapeComment, '(a, a, a, f6.1, a, f6.1, a, f5.1, a)') "source ", sourceShape(1:i-1), &
                      "(", sourceXPosArray(0),",", sourceYPosArray(0), ",", radiusDetectedArray(0), ")"            
               end if

               allocate(rateSourceArray(0:nSource-1))
               !---VNY 27.03.2002
               rateSourceArray=0.0
               
               rateSourceArray = real(iSourceArray)
               allocate(timeSource(0:nSource))
               timeStart = real64Attribute(sourceTable, "TSTART")
               timeDel = real64Attribute(sourceTable, "TIMEDEL")

               do i = 0, nSource
                  timeSource(i) = timeStart + dble(i-1)*timeDel
               enddo

               ! get a handle on the background rates file
               backSet = dataSet(TRIM(backRates), read)   

               !---VNY 17.01.2001 Protection against non existent block RATE: bkgRateExists
               bkgRateExists = .false.
               if ( hasBlock(backSet, "RATE")) then
                  bkgRateExists = .true.
               else
                  call warning("bkgRATE", "Block RATE does not exist in the background rates file")
               endif

               medianBkgError=1.0
               medianBkg=0.0
               backMax=0
               nBack = 0
               nRegion = 0
               if (bkgRateExists) then
                  backTable = table(backSet,"RATE")
                  nBack = numberOfRows(backTable)

                  ! ---VNY 26.03.2002 protection agains too big values of nBack
                  if (nBack .gt. 10000) nBack=9999
                  if (nBack .lt. 0) nBack=-nBack

                  rateBackCol = column(backTable,"COUNTS", read)
                  iBackArray => int32data(rateBackCol)
                  
                  ! Calculating the maximal background rate
                  backMax=0
                  do i=0, nBack-1
                     if (iBackArray(i) .gt. backMax) backMax=iBackArray(i)
                  enddo

                  ! Build the background distribution
                  allocate(backDist(1:backMax+1))
                  backDist=0
                  do i=0,nBack-1
                     do j=0,backMax
                        if (iBackArray(i) .eq. j) backDist(j+1)=backDist(j+1)+1
                     enddo
                  enddo

                  ! Estimate the background median
                  medianFound=.false.
                  backSum=0.0
                  backSumFound=0.0
                  j=0
                  do i=1,backMax+1
                     backSum=backSum+real(backDist(i))
                     if (real(nBack)/2.0 .le. backSum .and. .not. medianFound) then
                        medianFound=.true.
                        j=i
                        backSumFound=backSum
                     endif
                  enddo
                  medianBkg=1.0
                  if (backSumFound .gt. 0.0) medianBkg=real(j)*real(nBack)/2.0/backSumFound
                  medianBkgError=sqrt(medianBkg)
                  deallocate(backDist)

               endif ! bkgRateExists

               !--- VNY 14.11.2001: reading the background aperture radius detected by OMREGION
               !--- VNY 07.02.2002: using backRegSet in stead of backSet (for the case of absense of the last one)
               backRegionTable = table(backRegSet, "REGION")
               backShapeCol = column(backRegionTable, "SHAPE", read)
               backRadiusCol = column(backRegionTable, "R", read)
               backXPosCol = column(backRegionTable, "X", read)
               backYPosCol = column(backRegionTable, "Y", read)
               nBackRegion = numberOfRows(backRegionTable)
               backRadiusArray => real32data(backRadiusCol)
               backXPosArray => real32data(backXPosCol)
               backYPosArray => real32data(backYPosCol)

               nBackAvoid=0
               do ia=1,nBackRegion
                  backShape = stringCell(backShapeCol,ia-1)
                  if (TRIM(backShape) .eq. "!CIRCLE") then
                     nBackAvoid=nBackAvoid+1
                     if (nBackAvoid .le. 10) then
                        xBackAvoid(nBackAvoid)= backXPosArray((ia-1)*50)
                        yBackAvoid(nBackAvoid)=backYPosArray((ia-1)*50)
                        rBackAvoid(nBackAvoid)=backRadiusArray((ia-1)*50)
                     endif
                  endif
               end do
               if (nBackAvoid .gt. 10) nBackAvoid=10
               
               !---VNY 17.01.2002 Protection agains an empty background rates table
               if (nBack .lt. 1) then
                  nBack=0
                  bkgRateExists = .false.
                  call warning ("RATE", "Input background rates table is empty")
               endif

               !--- VNY 14.11.2001 
              ! Prepare comment on the background region
               backShapeComment=" "
               if (nBackRegion .gt. 0) then
                  backShape = stringCell(backShapeCol,0)
                  ! Calculate the background shape word length              
                  wordlength2: do i=1,20
                     if (ichar(backShape(i:i)) .eq. 0) then
                        exit wordlength2
                     end if
                  end do wordlength2
 
                  write(backShapeComment, '(a, a, a, f6.1, a, f6.1, a, f5.1, a, f5.1, a)') &
                       "background ", backShape(1:i-1), "(", backXPosArray(0),",", backYPosArray(0), &
                       ",", backRadiusArray(0), ",", backRadiusArray(1), ")"        
               end if

               bkgRateId = .true.
               if (bkgRateExists) then
                  allocate(rateBackArray(0:nBack-1))
                  rateBackArray = real(iBackArray)

                  allocate(timeBack(0:nBack))
                  bkgTimeStart = real64Attribute(backTable, "TSTART")
                  bkgTimeDel = real64Attribute(backTable, "TIMEDEL")

                  if (stringAttribute(backSet,"EXP_ID") /= &
                      stringAttribute(sourceSet,"EXP_ID") ) then
                     bkgRateId = .false.
                     call warning("diffExp","Source and background rates files from different exposures")
                  endif

                  if (stringAttribute(backRegSet,"EXP_ID") /= &
                      stringAttribute(backSet, "EXP_ID")) then
                     bkgRateId= .false.
                     call warning("diffExp", "Background rates and region files have different EXP_IDs")
                  endif

                  if (stringAttribute(sourceRegSet,"EXP_ID") /= &
                      stringAttribute(sourceSet, "EXP_ID")) then
                     bkgRateId = .false.
                     call warning("diffExp", "Source rates and region files have different EXP_IDs")
                  endif

                  do i = 0, nBack
                     timeBack(i) = bkgTimeStart + dble(i-1)*bkgTimeDel
                  enddo

                  ! the two time lines might be very different (if there is few or little background counts)
                  ! the master time line is taken from the source 
                  timeStart=timeSource(1)
                  timeEnd=timeSource(nSource)
               endif ! bkgRateExists
               allocate(timeOut(0:nSource-1))
               do i = 0, nSource-1
                  timeOut(i) = timeStart + dble(i)*timeDel
               enddo

               allocate(sourceArrayError(0:nSource-1))
               allocate(backArrayError(0:nBack-1))

               !---VNY 27.03.2002
               sourceArrayError=0.0
               backArrayError=0.0

               ! create an output dataset
               outSet = dataSet(outFile, create)

               call copyAllAttributes(attributable(sourceSet), attributable(outSet))
               call setAttribute(outSet, "CONTENT", "OM OSW SOURCE TIMESERIES")
               call setAttribute(outSet, "SRCNO", srcNo, "n", "Source number")

               !---15.05.2001 VNY: the output table title has been changed from RATES to RATE 
               outTable = addTable(outSet, "RATE", nSource)

               !---SRR (19/12/01) changed TIMEZERO value to 0.d0 (this should reflect any offset from the 
               !   TSTART parameter which gets propagated into the headers from obtstart I think). 
               call copyAllAttributes(attributable(sourceTable), attributable(outTable))
               call setAttribute(outTable, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
               call setAttribute(outTable, "TIMEDEL", timeDel, "s", "time binning interval")

               !---VNY 17.12.2001 Adding time column to the output table
               timeSourceOutCol = addColumn(outTable, "TIME", real64, "S")
               timeSourceOutArray => real64data(timeSourceOutCol)

               rateSourceOutCol = addColumn(outTable,"RATE", real32, "COUNTS/S")
               rateSourceOutArray => real32data(rateSourceOutCol)

               rateSourceErrOutCol = addColumn(outTable,"ERROR", real32,"COUNTS/S")
               rateSourceErrOutArray => real32data(rateSourceErrOutCol)

               rateBackOutCol = addColumn(outTable,"BACKV", real32, "COUNTS/S")
               rateBackOutArray => real32data(rateBackOutCol)

               rateBackErrOutCol = addColumn(outTable,"BACKE", real32, "COUNTS/S")
               rateBackErrOutArray => real32data(rateBackErrOutCol)

               ! VNY 07.02.2002 Using medianBkgError as the estimation for error bars when 
               ! source or background is zero.
               sourceArrayError=medianBkgError
               !---VNY 27.03.2002
               if (srcRateExists) then
                  do i=0,nSource-1
                     if (rateSourceArray(i) .gt. 0.0) sourceArrayError(i) = sqrt(rateSourceArray(i))
                  enddo
               endif

               ! VNY 07.02.2002 If there is no background counts at all, the background error bars are set to 0
               if (backMax .gt. 0) then
                  backArrayError=medianBkgError
               else
                  backArrayError=0.0
               endif

               if (bkgRateExists) then      
                  do i=0,nBack-1
                     if (rateBackArray(i) .gt. 0.0) backArrayError(i) = max(backArrayError(i), sqrt(rateBackArray(i)))
                  enddo
               endif ! bkgRateExists

               !---VNY 08.08.2001 Applying  the PSF-mapping function
               !******************************************************************************
               xPos0=dble(sourceXPosArray(0))
               yPos0=dble(sourceYPosArray(0))
               xRes=1
               yRes=1
               radiusDetected=radiusDetectedArray(0)
               ! VNY 12.02.2002 selection of radius for PSF-mapping
               psfRadius=max(photometryApertureRadius,radiusDetected, real(windowX1-windowX0+1),real(windowY1-windowY0+1))
               psfRadius=max(psfRadius, backRadiusArray(0), backRadiusArray(1))

!---VNY 16.04.2002 Test $$$$$$$$$$$$$$$$
!               detectedArea = PI * (radiusDetected+0.5) * (radiusDetected+0.5)
               detectedArea = PI * (radiusDetected+0.4) * (radiusDetected+0.4)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

               !---VNY 10.09.2001 Taking into account 0.5 pixel displacement of the coordinate system
               dxLeft=xPos0-windowX0 + 0.5 
               dxRight=windowX1-xPos0 + 0.5
               dyDown=yPos0-windowY0 + 0.5
               dyUp=windowY1-yPos0 + 0.5

               call CAL_omGetPSFmap(xPos0,yPos0, int(psfRadius), int(psfRadius), int(psfRadius), int(psfRadius), &
                   xRes, yRes, dble(countFrameRatio), psf)

!+++++++
!--- VNY 28.08.2001 Temporarely using a wider PSF : sqrt(((radius-6)**2)/100)
!              do iy=0,2*psfRadius
!                 do ix=0,2*psfRadius
!                    xc=real(ix-psfRadius)
!                    yc=real(iy-psfRadius)
!                    xyDistance=sqrt(xc*xc+yc*yc)
!                    if (abs(xyDistance) .le. 6.0) then 
!                       psf(ix,iy)=psf(ix,iy)+sqrt((xyDistance-6.0)*(xyDistance-6.0)/100.0)
!                    end if
!                 end do
!              end do
!--- VNY 28.08.2001 When the PSF will be corrected in CAL_omGetPSFmap, the above code should be removed
!++++++


               ! Calculation of the PSF fractions within different areas
               sumIn=0.  ! for PSF integrated over the area of the source inside the window 
               sumOut=0. ! the same, but outside the window
               sumIn2=0. ! PSF integrated within CALapertureRadius
               sumOut2=0. 
               sumIn3=0.
               sumOut3=0.
               sumBkg=0.0
               testSumBkg=0.0

               areaIn=0 ! Geometrical areas 
               areaOut=0
               areaIn2=0
               areaOut2=0
               areaIn3=0
               areaOut3=0
               areaBkg=0
               
               ! VNY 24.01.2002 Calculate the number of pixels in the background region
               ! that are not masked by the avoiding circle
               bkgExtractionArea=0
               do ix=windowX0, windowX1
                  do iy=windowY0,windowY1
                     outOfAvoid=.true.
                     do ia=1,nBackAvoid
                        dist=(real(ix)+0.5-xBackAvoid(ia))*(real(ix)+0.5-xBackAvoid(ia))+ &
                             (real(iy)+0.5-yBackAvoid(ia))*(real(iy)+0.5-yBackAvoid(ia))
                        dist=sqrt(dist)
                        if (dist .lt. rBackAvoid(ia)) then
                           outOfAvoid=.false.
                        endif
                     enddo
                     if (outOfAvoid) then
                        dist=(real(ix)+0.5-backXPosArray(0))*(real(ix)+0.5-backXPosArray(0)) +  & 
                             (real(iy)+0.5-backYPosArray(0))*(real(iy)+0.5-backYPosArray(0))
                        dist=sqrt(dist)
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$VNY check 16.04.2002
!---VNY 16.04.2002 Adding 0.5-pixel correctoion to the extraction radii
                        if (dist .gt. backRadiusArray(0)-0.4 .and. dist .le. backRadiusArray(1)) then
                           bkgExtractionArea=bkgExtractionArea+1
!                           testSumBkg=testSumBkg+psf(ix,iy)
                        endif
                     endif
                  enddo
               enddo

               ! Calculate psf-fractions inside and outside OSW 
               do iy=0,2*int(psfRadius)
                  do ix=0,2*int(psfRadius)
                     !---VNY 11.12.2001 Checking if ix and iy fall down to the avoiding circle
                     outsideOfAvoidingCircle = .true.
                     if (nAvoidingSources .gt. 0) then
                        ! Calculate distance from the avoiding centre
                        do i = 1, nAvoidingSources
                           xc=real(ix)+0.5-psfRadius-avoidX(i)
                           yc=real(iy)+0.5-psfRadius-avoidY(i)
                           xyDistance=sqrt(xc*xc+yc*yc)
                           if (xyDistance .lt. avoidR(i)) then
                              outsideOfAvoidingCircle = .false.                            
                           end if
                        end do
                     end if
                     if (nBackAvoid .gt. 0) then
                        do i=1,nBackAvoid
                           dist=(real(ix)+0.5-xBackAvoid(i))*(real(ix)+0.5-xBackAvoid(i))+ &
                                (real(iy)+0.5-yBackAvoid(i))*(real(iy)+0.5-yBackAvoid(i))
                           dist=sqrt(dist)
                           if (dist .lt. rBackAvoid(i)) then
                              outsideOfAvoidingCircle=.false.
                           endif
                        enddo
                     endif ! nBackAvoid .gt. 0

                     xc=real(ix)+0.5-psfRadius
                     yc=real(iy)+0.5-psfRadius
                     xyDistance=sqrt(xc*xc+yc*yc)
                     !---VNY 17.08.2001
                     ! For the frame limits (calculation of the psf fractions within  
                     ! and outside the OM science window)
!---VNY 16.04.2002 Adding 0.5-pixel correction to the radiusdetected
!                     if (xyDistance .le. radiusDetected+0.5) then 
                     if (xyDistance .le. radiusDetected+0.4) then 
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$         
                        if (xc .lt. dxRight .and. xc .gt. -dxLeft &
                             !  VNY 10.12.2001 An additional condition is necessary here in order to take into account 
                             !                 photon counts falling into the neighbour source avoiding area
                            .and. yc .lt. dyUp .and. yc .gt. -dyDown &
                            .and. outsideOfAvoidingCircle) then
                           sumIn=sumIn+psf(ix,iy)
                           areaIn=areaIn+1
                        else                          
                           sumOut=sumOut+psf(ix,iy)
                           areaOut=areaOut+1
                        end if
                     else
                        if ( xc .lt. dxRight .and. xc .gt. -dxLeft &
                             .and. yc .lt. dyUp .and. yc .gt. -dyDown &
!---VNY 16.04.2002 Adding -0.5-pixel correction to the backRadiusArray
                             .and. xyDistance .ge. backRadiusArray(0)-0.4 &
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
                             .and. xyDistance .lt. backRadiusArray(1) &
                             .and. outsideOfAvoidingCircle) then
                           sumBkg=sumBkg+psf(ix,iy)
                           areaBkg=areaBkg+1
                        endif
                     endif

                     ! calculation of the ratio between the cal-radius fraction of psf
                     ! and its photometry-radius fraction (sumIn2 and sumOut2)
                     if (xyDistance .lt. calApertureRadius) then
                        sumIn2=sumIn2+psf(ix,iy) ! For the PHOT to CAL ratio
                        areaIn2=areaIn2+1

                        ! Calculation of the ratio between the detected radius fraction of PSF
                        ! and CAL-radius fraction (sumIn3 and sumOut3)
!---VNY 15.04.2002 Correcting radiusDetected by 0.5-pixel 
!                        if (xyDistance .le. radiusDetected+0.5) then
                        if (xyDistance .le. radiusDetected+0.4) then
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
                           sumIn3=sumIn3+psf(ix,iy)
                           areaIn3=areaIn3+1
                        else
                           sumOut3=sumOut3+psf(ix,iy)
                           areaOut3=areaOut3+1
                        endif
                     else
!---VNY 15.04.2002 sumOut2 is related to photometryApertureRadius
                        if (xyDistance .lt. photometryApertureRadius) then
                           sumOut2=sumOut2+psf(ix,iy)
                           areaOut2=areaOut2+1
                        endif
                     end if
                  end do
               end do

               if (sumIn .gt. 0.0) then
                  ratioBS=sumBkg/sumIn
               else
                  ratioBS=0.0
               endif

               write(errstr,'(a,f6.1)') "Source extraction area (pixels): ",real(areaIn)
               call message(AppMsg, 2)

               write(errstr,'(a,f6.1)') "Background extraction area (pixels): ",bkgExtractionArea
               call message(AppMsg, 2)

               photcal=1.0 ! ratio between photometry-radius PSF fraction and its CAL-radius fraction
               caldet=1.0  ! ratio between CAL-radius PSF fraction and its detected-radius fraction
               countsRatio=1.0
               psfFrac=0.0
               if (sumIn .gt. 0.) then
                  countsRatio=(sumIn+sumOut)/sumIn ! to scale src counts to the CAL-circular area 
                  psfFrac=sumIn/(sumIn+sumOut)
               end if
               if (sumIn2 .gt. 0.) then
                  photcal=(sumIn2+sumOut2)/sumIn2 ! to scale src counts to the Photometry circular area
               endif
               if (sumIn3 .gt. 0.) then
                  caldet=(sumIn3+sumOut3)/sumIn3 ! to scale src counts from the detected to CAL circular areas
               endif

               mgAver=0.0
               mgMin=1.E30
               mgMax=1.E-30
               nmgAver=0.0
               rateAver=0.0
               
               !---VNY 27.03.2002
               if (srcRateExists) then

                  ! loop over the time line
                  do i = 0, nSource - 1

                     ! find the position of this master time in the background timeseries
                     ! and interpolate to get the background rate
                     if (bkgRateExists) then
                        call dlocate(timeBack, nBack, timeOut(i), j)
                        !---VNY 22.01.2002 protection against zero value for j
                        ja=0
                        if (j .lt. 0) then 
                           j=0
                           ja=1
                        endif

                        background=0.0

                        if (backMax .gt. 0) then
                           backErr=medianBkgError
                        else
                           backErr=0.0
                        endif

                        if (j .lt. nBack-1) then
                           prop = (timeOut(i)-timeBack(j)) / (timeBack(j+1)-timeBack(j))
                           background = rateBackArray(j) + &
                                prop * (rateBackArray(j+1) - rateBackArray(j))
                           if (background .lt. 0.0 .or. ja .eq. 1) background=0.0
                           if (background .gt. medianBkg) then 
                              backErr = BackArrayError(j) + &
                                   prop * (BackArrayError(j+1) - BackArrayError(j))
                           endif ! background .gt. medianBkg
                        else
                           if (j .lt. nBack) then
                              background= rateBackArray(j)
                              if (background .gt. medianBkg) then
                                 backErr=BackArrayError(j)
                              endif ! background .gt. medianBkg
                           else
                              background=0.0
                           endif
                        endif ! j .lt. nBack                     
                     else ! bkgRateExists
                        background=0.0
                        if (backMax .gt. 0) then
                           backErr=medianBkg
                        else
                           backErr=0.0
                        endif
                     endif ! bkgRateExists

                     ! VNY 17.12.2001 Adding output TIME column
                     timeSourceOutArray(i)=timeOut(i)
                     ! (background: background counts collected from the backArea)        

                     source=rateSourceArray(i)
                     sourceErr=SourceArrayError(i)
                                     
                     ! 13.11.2001 scaling background to the source area:
                     ! 24.01.2002 VNY Source counts are collected from the source extraction area, that is sumIn
                     ! and background counts are collected from the bkgExtractionArea
                     ! 13.02.2002 VNY scaling with the correction for PSF partly present in the bkg-area

!---VNY 11.04.2002 Solving a system of two equations
                     testBkg=0.0
                     if (sumIn .ne. 0.0) then
                        testZnam=areaBkg-areaIn*sumBkg/sumIn
                        if (testZnam .ne. 0.0) then
                           testBkg=(background-source*sumBkg/sumIn)/testZnam
                           testSrc=(source-testBkg*areaIn)/sumIn
!---VNY 15.04.2002 These two values are used if the result of solving two equations is taken into account
                           background=testBkg*areaIn
                           backErr=backErr*areaIn/bkgExtractionArea
!---
                        endif ! testZnam .ne. 0.0
                     endif ! sumIn .ne. 0.0

!---VNY 12.04.2002 These two values are used when background is calculated without the system of two equations
!                     background=background*real(areaIn)/bkgExtractionArea
!                     backErr=backErr*real(areaIn)/bkgExtractionArea

                     !--- VNY 31.10.2001 Now  the source counts could be corrected for asymmetrical position 
                     !    of the source within the OSW (ratio between the PSF fraction within the OSW
                     !    and outside of it), then the background counts should be scaled to the standard photometry radius
                     
                     ! 13.11.2001 Subtracting background from the source counts in order to correct the source PSF
                     bgsubtracted=0 ! Indicator of whether background was subtracted or not

                     source = source - background  

                     bgsubtracted=1


                     ! 13.11.2001 The source counts with background substracted are corrected for the PSF
                     !            fraction outside OSW
!---VNY 11.04.2002 Do this only if background was substracted
                     source = source * countsRatio
                     sourceErr = sourceErr * countsRatio
                     !Now the source is normalised to the circle area with the detected radius: radiusDetectedArray(0)
                     !  Converting counts to rates, dividing by time bins
                     if (timeDel .gt. 0) then
                        background = background / timeDel
                        backErr = backErr / timeDel
                        source = source / timeDel
                        sourceErr = sourceErr / timeDel
                     end if

                     ! Now source has counts collected during timeDel from the area limited by radiusDetected
                     ! The background has counts collected during timeDel from source extraction area (areaIn)
                     
                     ! 24.01.2002 VNY Scaling background to the sourceArea (which is areaIn+areaOut or sourceArea)
                     if (areaIn .gt. 0) then
                        background=background*(real(areaIn)+real(areaOut))/real(areaIn)
                        backErr=backErr*(real(areaIn)+real(areaOut))/real(areaIn)
                     endif
                     ! Now background is scaled to the sourceArea

                     !--- here source could be corrected for the coincidence loss
                     ! 13.11.2001 before the coincidence loss correction source and background are to be scaled 
                     !            to the standard calApertureRadius
!--- VNY 16.04.2002
                     if (radiusDetectedArray(0)+0.4 .gt. 0.) then
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
                        corrRate = calApertureRadius * calApertureRadius 
                        corrRate = corrRate / ((radiusDetectedArray(0)+0.4) * (radiusDetectedArray(0)+0.4))
!$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

                        source = source * caldet ! Scaling source to the calApertureRadius
                        sourceErr = sourceErr * caldet

                        background = background * corrRate ! Scaling background to the calApertureRadius
                        backErr = backErr * corrRate
                     end if

                     ! 13.11.2001 returning background to the source in order to be able to apply
                     !  the coincidence loss correction 

                     source = source + background

!---VNY 17.04.2002 The source errors are not increasing here, because we are restoring previously 
!  subtracted background (it was just scaling of the source counts for the missed photons outside 
!  the window)

                     !******************************************************
                     ! Need to correct the rates for coincident losses, etc-
                     !******************************************************
                     !---VNY 25.10.2001 initialisation of corrRateErrMinus1 and corrRateErrPlus1
                     corrRateErrMinus1=0.0
                     corrRateErrPlus1=0.0

                     if (background .ge. 0.0) then   
                        corrRate = CorrectedCountRate(dble(background), backErr, frameTime, deadFrac,  &
                             extend, corrRateErrMinus1, corrRateErrPlus1)
                        background = corrRate
                     else
                        nNegative=nNegative+1
                     endif ! background
 
                     !---VNY 25.10.2001 Initialisation of corrRateErrMinus and corrRateErrPlus
                     corrRateErrMinus=0.0
                     corrRateErrPlus=0.0

                     if (source .ge. 0.0) then
                        ! VNY 13.08.2001 Now the correction for the coincidence loss and dead fraction is made before
                        ! scaling the count rates to the photometric aperture radius
                        corrRate = CorrectedCountRate(dble(source), sourceErr, frameTime, deadFrac, &
                             extend, corrRateErrMinus, corrRateErrPlus)
                        source = corrRate
                     else     
                        nNegative=nNegative+1
                     endif ! Source .gt. 0.0
                     rateBackOutArray(i)=background
                     rateBackErrOutArray(i) = real(abs(max(corrRateErrMinus1, corrRateErrPlus1)))     
                     rateSourceOutArray(i) = source -  rateBackOutArray(i)
!---VNY 17.04.2002 Increasing the source error bars by about 2% because of a systematic non-poissonian
!  error for bright sources  after the coincidence loss correction 
                     rateSourceErrOutArray(i) = sqrt(real(abs(max(corrRateErrMinus, corrRateErrPlus)))**2 &
                          + rateBackErrOutArray(i)**2 + (0.02 * rateSourceOutArray(i))**2)

                     !--- VNY 26.06.2001 Calculation of magnitudes
                     mag(iLoop)=1.e30
                     magErr(iLoop)=1.e30               
                     if (rateSourceOutArray(i) .gt. 0.0) then
                        corrRate=rateSourceOutArray(i)
                        ! VNY 3.12.2001 Scaling source counts to the photometry aperture radius 
                        corrRate = corrRate * photcal
                        call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                             corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                        mag(iLoop)=real(corrMag)
                        magErr(iLoop)=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
                        mgAver=mgAver+mag(iLoop)
                        nmgAver=nmgAver+1.0
                        rateAver=rateAver+rateSourceOutArray(i)
                        if (mag(iLoop) .lt. mgMin) then
                           mgMin=mag(iLoop)
                        end if
                        
                        if (mag(iLoop) .gt. mgMax) then
                           mgMax=mag(iLoop)
                        end if
                     else   
                        nNegative=nNegative+1
                     endif ! rateSourceOutArray
            
                     !--- VNY 26.06.2001 Calculation of background magnitude
                     mag(iLoop)=1.e30
                     magErr(iLoop)=1.e30
                     if (rateBackOutArray(i) .gt. 0.0) then
                        corrRate=rateBackOutArray(i)
                        ! VNY 3.12.2001 Scaling background rates to the photometry aperture radius
                        if (calApertureRadius .gt. 0.0) then
                           corrRate = corrRate * (photometryApertureRadius/calApertureRadius)**2
                        endif
                        call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
                             corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
                        mag(iLoop)=real(corrMag)
                        magErr(iLoop)=real(abs(max(corrMagErrMinus, corrMagErrPlus)))
                     endif ! ratebackOutArray
                  enddo ! loop over all sources (i)

                  if (nmgAver .gt. 0.0) then
                     mgAver=mgAver/nmgAver
                     rateAver = rateAver/nmgAver
                  end if

                  !---VNY 21.09.2001 Median filtration of the source rates to make a warning 
                  do i = 0, nSource - 10
                     ! Array for the median filter
                     do ix= 1, 9
                        rateAuxArray(ix)=rateSourceOutArray(i+ix)
                     end do
                     ! Ordering the auxiliar array
                     do ix=1,8
                        do iy=ix+1,9
                           if (rateAuxArray(iy) .gt. rateAuxArray(ix)) then
                              rateAux=rateAuxArray(ix)
                              rateAuxArray(ix)=rateAuxArray(iy)
                              rateAuxArray(iy)=rateAux
                           end if
                        end do
                     end do
                     rateAux=rateAuxArray(5)
                     if (rateSourceOutArray(i) .gt. rateAux*30.0) then
                        call warning("sourceRate", "High value of the source rate")
                     end if
                  end do
                  ! Repeat analysis for the last 9 values
                  do i=nSource-9, nSource-1
                     if (rateSourceOutArray(i) .gt. rateAux*30.0) then
                        call warning("sourceRate", "High value of the source rate")
                     end if
                  end do
               endif ! srcRateExists

               psfApp= .true.
               vignApp= .false.
               backApp= .true.
               deadApp= .true.

               !---VNY 30.08.2001 Three new attributes have been added to the output table: MG_AVER, MG_MIN, MG_MAX
               call setAttribute(outTable, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
               call setAttribute(outTable, "MAGNITUD", mgAver, "mg", "Average magnitude of the source")
               call setAttribute(outTable, "MAGN_MIN", mgMin, "mg", "Minimum magnitude in this time series")
               call setAttribute(outTable, "MAGN_MAX", mgMax, "mg", "Maximum magnitude in this time series")

               !---VNY 19.11.2001 Attribute with the PSF-fraction within the OSW
               call setAttribute(outTable, "PSF_FRAC", psfFrac, " ", "Fraction of PSF inside OSW")
               call setAttribute(outTable, "VIGNET", vignet ,"Geometrical vignetting of the source")
               call setAttribute(outTable, "NPIXSOU", sourceArea, "pix", "Source area")
               call setAttribute(outTable, "NPIXBACK", backArea, "pix", "Background area")
               call setAttribute(outTable, "SRRADIUS", radiusDetectedArray(0), "pix", "Source extraction radius")
               call setAttribute(outTable, "CLRADIUS", calApertureRadius, "pix", "Coincidence loss correction radius")
               call setAttribute(outTable, "PHRADIUS", photometryApertureRadius,"pix", "Photometry aperture radius")
               call setAttribute(outTable, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
               call setAttribute(outTable, "DEADAPP", deadApp, " Dead time correction applied")
               call setAttribute(outTable, "BACKAPP", backApp,  "Background is subtracted")
               call setAttribute(outTable, "PSF_CAPP", psfApp, "PSF-fraction correction applied")
               !---VNY 19.11.2001
           
               call setAttribute(outTable, "HDUCLAS1", "LIGHTCURVE", "Light curve series")
               call setAttribute(outTable, "HDUCLAS2", "NET", "Background subtracted")
               call setAttribute(outTable, "HDUCLAS3", "RATE", "Exposure corrected")
               call setAttribute(outTable, "RA", sourceRA(0), "deg", "Right Ascension of the source")
               call setAttribute(outTable, "DEC", sourceDEC(0), "deg",  "Declination of the source")
               call setAttribute(outTable, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
               call setAttribute(outTable, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
               call setAttribute(outTable, "QFLAG", sourceQF(0), " ",  "Quality flag")
               call setAttribute(outTable, "CFLAG", sourceCF(0), " ",  "Confusion flag")
               call addComment(outTable, " ")
               call addComment(outTable,  sourceShapeComment)
               call addComment(outTable, backShapeComment)

               if (minval(rateSourceOutArray) < 0.) then
                  call warning("negVals", "Background subtracted rate has negative values")
               endif

               ! release the dal handles
!---VNY 22.03.2002
               call release(rateBackErrOutCol)
               call release(rateBackOutCol)
               call release(rateSourceErrOutCol)
               call release(rateSourceOutCol)
               call release(timeSourceOutCol)
               call release(outTable)
               call release(outSet)
!---VNY 22.03.2002
               call release(backYPosCol)
               call release(backXPosCol)
               call release(backRadiusCol)
               call release(backShapeCol)
               call release(backRegionTable)
!---VNY 28.03.2002
               if (bkgRateExists) then
                  call release(rateBackCol)
                  call release(backTable)
               endif
               call release(backSet)
            else ! hasBlock(sourceSet, "RATE")
               call warning("Rate", "Block RATE does not exist in the source rates file")
            endif ! hasBlock(sourceSet, "RATE")
!---VNY 22.03.2002
            call release(sourceYPosCol)
            call release(sourceXPosCol)
            call release(sourceRadiusCol)
            call release(sourceShapeCol)
            call release(regionTable)
            call release(rateSourceCol)
            call release(sourceTable)
            call release(sourceSet)

            deallocate(rateSourceArray)
            deallocate(timeSource)
            if (bkgRateExists) then
               deallocate(rateBackArray)
               deallocate(timeBack)
            endif
            deallocate(timeOut)
            deallocate(sourceArrayError)
            deallocate(backArrayError)         
            !*************************
            ! Release handles on data
            !*************************
            do i=1, noColumns
               call release(inColumnH(i))
            end do
         endif ! nSources>0
         call release(inTableH)
      else ! hasBlock(inDataSetH)
         !************************************
         ! Incompatible file - can't continue
         !************************************
         write (errstr, *) "Could not find any compatibile blocks in set ", sourceList
         call message(AppMsg, 2)
      endif ! hasBlock(inDataSetH)
      call release(inDataSetH)
      call release(sourceRegSet)
      call release(backRegSet)
    end subroutine omlcbuild

!*********************************************************
! Function CorrectedCountRate
! Corrects the raw count rate for coincidence losses, etc  
!*********************************************************
real(kind=double) function CorrectedCountRate(rawCountRate, & 
                  countRateError, frameTime, deadFrac, extended, &
                  countRateErrorMinus, countRateErrorPlus)
    real(kind=double), intent(in)    :: rawCountRate
!---VNY 29.03.2002 single type replaced with double (countRateError)
    real(kind=double), intent(in) :: countRateError
    real(kind=double), intent(out)    :: countRateErrorMinus
    real(kind=double), intent(out)    :: countRateErrorPlus
 
    real(kind=double), intent(in) :: frameTime
    real(kind=double), intent(in) :: deadFrac
    logical, intent(in) :: extended
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate

    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))

    correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0d0, extended)

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************

    if (CorrectedRate%value .lt. 1e10) then
        CorrectedCountRate = CorrectedRate%value
        countRateErrorMinus = correctedRate%errorMinus
        countRateErrorPlus = correctedRate%errorPlus

    else
       !***************************************************************
       !---VNY 01.11.2001 For the case of CAL_omPhotoNatural would fail
       !     the count rate will not be corrected
       !***************************************************************
       if (rawCountRate .lt. 1e10) then
          CorrectedCountRate = rawCountRate
          countRateErrorMinus = countRateError
          countRateErrorPlus = countRateError
       else
          CorrectedCountRate = -1.0d0
          countRateErrorMinus = -1.0d0
          countRateErrorPlus = -1.0d0
       end if
    end if

end function CorrectedCountRate


subroutine dlocate(xx, n, x, j)
! VNY 25.01.2002; If the dimension of the array xx is 1 or 0
! this subroutin can crash due to subscript. Protection 
! against this crash has been added
   integer(kind=int32), intent(in) :: n
   real(kind=double)   :: xx(0 : n - 1)
  
   real(kind=double)   :: x
   integer(kind=int32) :: j

   integer(kind=int32) :: ju
   integer(kind=int32) :: jl
   integer(kind=int32) :: jm

   jl = 0
   ju = n + 1

! VNY 25.01.2002 Protection against one-element array
   if (n .le. 1) then
      j = 0
      return
   endif

   if (x .ge. xx(n - 1)) then
       j = n - 1
       return
   end if 
      
  

   if (x .le. xx(0)) then
      j = 0
      return
   endif

   do while(ju - jl > 1)
      jm = (ju + jl) / 2

      if (x .gt. xx(jm - 1)) then
         jl = jm
      else
         ju = jm
      end if
   end do

   j = max(jl - 1, 0)
   
end subroutine dlocate

   
end module omlcbuild_mod
